'''
easycbc256: 对 AES-CBC-256 加密算法的二次封装, 使得 加密/解密 流程更加简单.

该数据结构线程安全.

easycbc256 具有以下特点:
    1. 自动添加和管理iv
    2. 自动添加和管理salt
    3. 安全性较原生 AES-CBC-256 未降低.

注意: 使用 easycbc256 加密得到的数据, 只能使用 easycbc256 解密 !!!

Note: the data encrypted with easycbc256 can only be decrypted with easycbc256 !!!
'''


import hashlib
from random import random as r_random
from Crypto.Cipher import AES  # pip install pycryptodome
from base64 import encodebytes as bytes_into_base64  # 允许其它模块导入
from base64 import b64decode as base64_into_bytes  # 允许其它模块导入


class easycbc256():

    def __init__(self, allow_empty_key=True, fastkey=False):
        self.__allow_empty_key = allow_empty_key
        if fastkey is not False:
            self.__fastkey = self.__check_key(fastkey)

    def __random_bytes(self, size=16) -> bytes:
        # 不使用 os.urandom(16) , 因为它效率较低.
        rb = hashlib.shake_256(str(r_random()).encode('utf8')).digest(size)
        return rb

    def __padding(self, ptext) -> bytes:
        if type(ptext) is str:
            ptext = ptext.encode('utf8')
        dsize = 16 - len(ptext) % 16
        ptext += chr(dsize).encode('utf8') * dsize
        return ptext

    def __check_key(self, key):
        if key:
            if type(key) is str:
                key = key.encode('utf8')
            elif type(key) is not bytes:
                raise Exception('Nonempty key must be an instance of str or bytes')
        elif not self.__allow_empty_key:
            raise Exception('Key cannot be empty')
        elif key is not None:
            raise Exception('An empty key can only be None')
        return key

    # 加密
    def encrypt(self, key=None, ptext='') -> bytes:
        key = self.__check_key(key)
        iv = self.__random_bytes(16)
        salt = self.__random_bytes(32)
        cbckey = hashlib.shake_256(key + salt).digest(32)
        # AES密钥支持128位、192位、256位, 此处 32字节 * 8 = 256位
        # 由于密钥经过 shake_256 转换, 破解难度为: min(256 ** 32, 2 ** 256)
        ptext = self.__padding(ptext)
        ctext = AES.new(cbckey, AES.MODE_CBC, iv=iv).encrypt(ptext)
        ctext = salt + iv + ctext
        return ctext

    def __ctext_parser_for_decrypt(self, ctext):
        salt, iv, ctext = ctext[:32], ctext[32:48], ctext[48:]
        return salt, iv, ctext
    
    # 解密
    def decrypt(self, key=None, ctext=b'') -> bytes:
        key = self.__check_key(key)
        salt, iv, ctext = self.__ctext_parser_for_decrypt(ctext)
        cbckey = hashlib.shake_256(key + salt).digest(32)
        ptext = AES.new(cbckey, AES.MODE_CBC, iv=iv).decrypt(ctext)
        ptext = ptext[:-ptext[-1]]  # self.__padding 的逆过程
        return ptext

    # 快速加密
    def fasten(self, ptext='') -> bytes:
        iv = self.__random_bytes(16)
        salt = self.__random_bytes(32)
        cbckey = hashlib.shake_256(self.__fastkey + salt).digest(32)
        # AES密钥支持128位、192位、256位, 此处 32字节 * 8 = 256位
        # 由于密钥经过 shake_256 转换, 破解难度为: min(256 ** 32, 2 ** 256)
        ptext = self.__padding(ptext)
        ctext = AES.new(cbckey, AES.MODE_CBC, iv=iv).encrypt(ptext)
        ctext = salt + iv + ctext
        return ctext

    # 快速解密
    def fastde(self, ctext=b'') -> bytes:
        salt, iv, ctext = self.__ctext_parser_for_decrypt(ctext)
        cbckey = hashlib.shake_256(self.__fastkey + salt).digest(32)
        ptext = AES.new(cbckey, AES.MODE_CBC, iv=iv).decrypt(ctext)
        ptext = ptext[:-ptext[-1]]  # self.__padding 的逆过程
        return ptext


if __name__ == '__main__':

    cbc = easycbc256()
    key = '枫桥夜泊'
    ptext = '姑苏城外寒山寺，夜半钟声到客船。'
    ctext = cbc.encrypt(key, ptext)
    p2 = cbc.decrypt(key, ctext).decode('utf8')

    print('test:\n')
    print(f"[key] {key}\n")
    print(f"[ptext] {ptext}\n")
    print(f"[ctext] {ctext}\n")
    print(f"[p2] {p2}\n")
    print(f"[ptext == p2] {ptext==p2}\n")
