'''
@author: jutooy

Expiredict: 一种类似字典, 但支持key自动过期的数据结构.

该数据结构线程安全.

工作原理:
    1. 赋值一个新的 key:value 时, 将根据 clear_interval(时区的宽度) 计算"该key的失效时间属于哪个时区", 把此key放入对应时区的容器中.
    2. expiredict.__auto_clear 将定期自动检查内存中的 key 是否过期, 但它不会检查所有的key, 而只检查已过期的时区内的key.

其它功能:
    1. 支持对已存在的 "key:value" 键值对重置剩余生命.
    2. 支持对已存在的 "key:value" 键值对 增加/减少 生命.
'''


import asyncio
from time import time as nowtime
from .varlock import varlock


class KEYERR(): ...


class expiredict():
    
    def __init__(self, clear_interval=None, save_interval=None, savefunc=None):
        '''
        clear_interval: 自动清理间隔, 单位: s
        savefunc:
            自定义的持久化函数
            若启用自动保存组件, 则需传入此函数, 且该函数必须接收且仅接收"item"参数.
            若启用自动保存组件且未传入此函数, 则将调用expiredict内置的save方法, 该方法无任何实质性行为, 而仅仅打印一些提示语.
        '''
        self.__start_time = int(nowtime())
        self.__clear_interval = clear_interval
        self.__save_interval = save_interval
        self.__zone = {float('inf'): []}
        self.__item =  {}
        self.__lock = varlock()

        if not clear_interval:
            self.add = self.__add_without_autoclear
        elif clear_interval > 0:
            asyncio.create_task(self.__auto_clear())  # 启动自动清理组件
        else:
            raise Exception('clear_interval must be greater than 0')

        if save_interval:
            if save_interval < 0: raise Exception('save_interval must be greater than 0')
            if savefunc: self.save = savefunc
            asyncio.create_task(self.__auto_save())  # 启动自动保存组件


    ''' 基础用法 '''


    async def add(self, key, value, life=0, endtime=0):
        '''
        add: 赋值
        life: 存活时长, 该值会自动转换成 endtime
        endtime: 失效时间
        life 与 endtime 只能有1个真值, 否则将自动报错
        '''
        async with self.__lock:
            if life and endtime:
                raise Exception('please choose between life and endtime')
            elif life:
                if life > 0:
                    endtime = nowtime() + life
                    zonepk = int((endtime - self.__start_time) / self.__clear_interval)  # 判断失效时间处于哪个时间区间
                    self.__item[key] = [value, endtime]
                    self.__zone.setdefault(zonepk, []).append(key)
                else:
                    self.__item.pop(key, None)
            elif endtime:
                if endtime > nowtime():
                    zonepk = int((endtime - self.__start_time) / self.__clear_interval)  # 判断失效时间处于哪个时间区间
                    self.__zone.setdefault(zonepk, []).append(key)
                    self.__item[key] = [value, endtime]
                else:
                    self.__item.pop(key, None)
            else:
                self.__item[key] = [value, 0]
        return True

    async def __auto_clear(self):
        item = self.__item
        zone = self.__zone
        while True:
            minkey = min(zone)
            now = nowtime()
            zonepk = int((now - self.__start_time) / self.__clear_interval)  # 判断当前处于哪个时间区间
            # zonepk = int(...) 这里使用 int() 而非 ceil() 的原因: 使每个时间区间都等长
            if minkey < zonepk:
                async with self.__lock:
                    for key in zone.pop(minkey):
                        value, endtime = item.get(key) or [None, 0]
                        if 0 < endtime <= now:
                        # 第2个比较符用 <= 而不用 < 的原因: 该 key:value 显然已经没有实际用处了, 因此不必保留到下一轮__auto_clear
                            item.pop(key, None)
            else:
                t = (zonepk + 1) * self.__clear_interval + self.__start_time - nowtime()
                if t > 0:
                    await asyncio.sleep(t)

    async def __add_without_autoclear(self, key, value, life=0, endtime=0):
        async with self.__lock:
            if life and endtime:
                raise Exception('please choose between life and endtime')
            elif life:
                if life > 0:
                    self.__item[key] = [value, nowtime() + life]
                else:
                    self.__item.pop(key, None)
            elif endtime:
                if endtime > nowtime():
                    self.__item[key] = [value, endtime]
                else:
                    self.__item.pop(key, None)
            else:
                self.__item[key] = [value, 0]
        return True

    async def fetch(self, key):
        value, endtime = self.__item[key]
        if 0 < endtime < nowtime():  # 这里用 endtime < nowtime() 而非 endtime <= nowtime()
            async with self.__lock:
                value, endtime = self.__item[key]
                if 0 < endtime < nowtime():
                    self.__item.pop(key)
                    raise KeyError('the key has expired')
        return value

    async def get(self, key, default=None):
        try:
            return await self.fetch(key)
        except:
            return default

    async def pop(self, key, default=KEYERR):
        try:
            async with self.__lock:
                value, endtime = self.__item.pop(key)
            if 0 < endtime < nowtime():  # 这里用 endtime < nowtime() 而非 endtime <= nowtime()
                raise KeyError('the key has expired')
            return value
        except Exception as e:
            if default is KEYERR:
                raise e
            return default

    async def items(self):
        for key in list(self.__item):
            try:
                yield key, await self.fetch(key)
            except:
                pass

    async def keys(self):
        async for key, value in self.items():
            yield key

    async def values(self):
        async for key, value in self.items():
            yield value

    async def len(self):
        return len([x async for x in self.items()])


    ''' 中阶用法, 使用需一定的领悟力 '''


    @staticmethod
    async def save(item):
        '''
        占位方法, 若启用自动保存组件, 则须在实例化 expiredict 时传入 savefunc参数 以覆盖此方法,
        且传入的 savefunc函数 必须接收且仅接收"item"参数.
        每隔 self.__save_interval 秒将自动异步执行此方法.
        '''
        print('Method "save" is being executed now.')

    async def __auto_save(self):
        while True:
            await asyncio.sleep(self.__save_interval)
            await self.save(dict([x async for x in self.items()]))

    def __getitem_without_pop(self, key):
        value, endtime = self.__item[key]
        if 0 < endtime < nowtime():
            raise KeyError('the key has expired')
        else:
            return value

    async def setlife(self, key, default=KEYERR, life=0, endtime=0):
        '''
        功能: 为key重置生命
        当key存在时, 为其重置生命, 并返回value.
        当key不存在时, 若传递了default, 则返回default并为其设置生命; 若未传递default, 则报错.
        '''
        if default is KEYERR:
            value = await self.fetch(key)
            await self.add(key, value, life, endtime)
            return value
        else:
            try:
                value = self.__getitem_without_pop(key)
                await self.add(key, value, life, endtime)
                return value
            except:
                await self.add(key, default, life, endtime)
                return default


    ''' 高阶用法, 使用需较高的领悟力 '''


    async def setlife_when_succeeded(self, key, default=KEYERR, life=0, endtime=0):
        '''
        当key存在时, 为其重置生命, 并返回value.
        当key不存在时, 若传递了default, 则返回default但不为其设置生命, 若未传递default, 则报错.
        '''
        try:
            value = await self.fetch(key)
            await self.add(key, value, life, endtime)
            return value
        except Exception as e:
            if default is KEYERR:
                raise e
            else:
                return default

    async def setlife_when_failed(self, key, default=None, life=0, endtime=0):
        '''
        当key存在时, 返回value但不为其重置生命.
        当key不存在时, 返回default并为其设置生命.
        '''
        try:
            return self.__getitem_without_pop(key)
        except:
            await self.add(key, default, life, endtime)
            return default

    async def inclife(self, key, default=KEYERR, clife=0, fullest=0):
        '''
        功能: 为key增加存活时长
        clife: 要增加多少时长, 正数表示增加, 负数表示减少, 0表示既不增加也不减少.
        fullest: 如果key的生命时长超过fullest, 就调整为fullest, 表示"本次"调整生命时长的上限, 0表示不做限制.
        default: 当key不存在时, 若传递了default, 则以default代替value, 若未传递default, 则报错.
        '''
        if fullest and fullest < 0:
            raise Exception('fullest must be greater than 0')
        elif default is KEYERR:
            value, endtime = self.__item[key]
            if 0 < endtime < nowtime():
                async with self.__lock:
                    value, endtime = self.__item[key]
                    if 0 < endtime < nowtime():
                        self.__item.pop(key)
                        raise KeyError('the key has expired')
            if endtime:
                endtime += clife or 0
                if fullest: endtime = min(nowtime() + fullest, endtime)
                if not endtime: endtime = -1  # 避免 entime == 0 被 self.add 视为无限生命
                await self.add(key, value, 0, endtime)
            elif fullest:
                await self.add(key, value, fullest, 0)
            return value
        else:
            try:
                return await self.inclife(key, clife, KEYERR, fullest)
            except:
                if clife and fullest:
                    await self.add(key, default, min(clife, fullest), 0)
                elif clife or fullest:
                    await self.add(key, default, clife or fullest, 0)
                else:
                    await self.add(key, default, 0, 0)
                return default
